@@ -42,7 +42,7 @@ File objects can't be overwritten:
 True
 >>> try:
 ...   filesystem.CreateFile(pathname)
-... except IOError, e:
+... except IOError as e:
 ...   assert e.errno == errno.EEXIST, 'unexpected errno: %d' % e.errno
 ...   assert e.strerror == 'File already exists in fake filesystem'
 
@@ -84,7 +84,6 @@ True
 True
 """
 
-import cStringIO
 import errno
 import heapq
 import os
@@ -92,17 +91,21 @@ import stat
 import sys
 import time
 import warnings
+try:
+  import cStringIO as io  # pylint: disable-msg=C6204
+except ImportError:
+  import io  # pylint: disable-msg=C6204
 
 __pychecker__ = 'no-reimportself'
 
-__version__ = '1.2'
+__version__ = '2.0'
 
-PERM_READ = 0400      # Read permission bit.
-PERM_WRITE = 0200     # Write permission bit.
-PERM_EXE = 0100       # Write permission bit.
-PERM_DEF = 0777       # Default permission bits.
-PERM_DEF_FILE = 0666  # Default permission bits (regular file)
-PERM_ALL = 07777      # All permission bits.
+PERM_READ = 0o400      # Read permission bit.
+PERM_WRITE = 0o200     # Write permission bit.
+PERM_EXE = 0o100       # Write permission bit.
+PERM_DEF = 0o777       # Default permission bits.
+PERM_DEF_FILE = 0o666  # Default permission bits (regular file)
+PERM_ALL = 0o7777      # All permission bits.
 
 _OPEN_MODE_MAP = {
     # mode name:(file must exist, need read, need write,
@@ -277,7 +280,7 @@ class FakeDirectory(FakeFile):
 
     Args:
       name:  name of the file/directory, without parent path information
-      perm_bits: permission bits. defaults to 0777.
+      perm_bits: permission bits. defaults to 0o777.
     """
     FakeFile.__init__(self, name, stat.S_IFDIR | perm_bits, {})
 
@@ -335,7 +338,7 @@ class FakeFilesystem(object):
     self.root = FakeDirectory(self.path_separator)
     self.cwd = self.root.name
     # We can't query the current value without changing it:
-    self.umask = os.umask(022)
+    self.umask = os.umask(0o22)
     os.umask(self.umask)
     # A list of open file objects. Their position in the list is their
     # file descriptor number
@@ -382,6 +385,26 @@ class FakeFilesystem(object):
     self.open_files[file_obj.filedes] = None
     heapq.heappush(self.free_fd_heap, file_obj.filedes)
 
+  def GetOpenFile(self, file_des):
+    """Returns an open file.
+
+    Args:
+      file_des:  file descriptor of the open file.
+
+    Raises:
+      OSError: an invalid file descriptor.
+      TypeError: filedes is not an integer.
+
+    Returns:
+      Open file object.
+    """
+    if not isinstance(file_des, int):
+      raise TypeError('an integer is required')
+    if (file_des >= len(self.open_files) or
+        self.open_files[file_des] is None):
+      raise OSError(errno.EBADF, 'Bad file descriptor', file_des)
+    return self.open_files[file_des]
+
   def CollapsePath(self, path):
     """Mimics os.path.normpath using the specified path_separator.
 
@@ -480,7 +503,7 @@ class FakeFilesystem(object):
     for this FakeFilesystem.
 
     Args:
-      paths:  (str) Zero or more paths to join.
+      *paths:  (str) Zero or more paths to join.
 
     Returns:
       (str) The paths joined by the path separator, starting with the last
@@ -1053,7 +1076,7 @@ class FakePathModule(object):
     """Returns the mtime of the file."""
     try:
       file_obj = self.filesystem.GetObject(path)
-    except IOError, e:
+    except IOError as e:
       raise OSError(errno.ENOENT, str(e))
     return file_obj.st_mtime
 
@@ -1091,14 +1114,31 @@ class FakeOsModule(object):
       warnings.warn(FAKE_PATH_MODULE_DEPRECATION, DeprecationWarning,
                     stacklevel=2)
       self.path = os_path_module
+    if sys.version_info < (3, 0):
+      self.fdopen = self._fdopen_ver2
+    else:
+      self.fdopen = self._fdopen
 
-  def fdopen(self, file_des, mode=None, bufsize=None):
+  def _fdopen(self, *args, **kwargs):
+    """Redirector to open() builtin function.
+
+    Args:
+      *args: pass through args
+      **kwargs: pass through kwargs
+
+    Returns:
+      File object corresponding to file_des.
+
+    Raises:
+      TypeError: if file descriptor is not an integer.
+    """
+    if not isinstance(args[0], int):
+      raise TypeError('an integer is required')
+    return FakeFileOpen(self.filesystem)(*args, **kwargs)
+
+  def _fdopen_ver2(self, file_des, mode='r', bufsize=None):
     """Returns an open file object connected to the file descriptor file_des.
 
-    WARNING:  This implementation currently returns the original
-      FakeFileWrapper, rather than creating a new one, which differs from the
-      standard fdopen() spec. If you need this behavior, please implement it.
-
     Args:
       file_des: An integer file descriptor for the file object requested.
       mode: additional file flags. Currently checks to see if the mode matches
@@ -1111,30 +1151,14 @@ class FakeOsModule(object):
     Raises:
       OSError: if bad file descriptor or incompatible mode is given.
       TypeError: if file descriptor is not an integer.
-      ValueError: if invalid mode is given.
     """
-    if not isinstance(file_des, (int, long)):
+    if not isinstance(file_des, int):
       raise TypeError('an integer is required')
-    if (file_des >= len(self.filesystem.open_files) or
-        self.filesystem.open_files[file_des] is None):
-      raise OSError(errno.EBADF, 'Bad file descriptor', file_des)
-    file_obj = self.filesystem.open_files[file_des]
 
-    if mode:
-      orig_mode = mode  # Save original mode for error messages.
-      # Normalize mode. Ignore 't', 'b', and 'U'.
-      mode = mode.replace('t', '').replace('b', '')
-      mode = mode.replace('rU', 'r').replace('U', 'r')
-
-      if mode not in _OPEN_MODE_MAP:
-        raise ValueError('Invalid mode: %r' % orig_mode)
-      must_exist, need_read, need_write, truncate, append = _OPEN_MODE_MAP[mode]
-      st_mode = file_obj.GetObject().st_mode
-      if ((need_read and not st_mode & PERM_READ) or
-          (need_write and not st_mode & PERM_WRITE)):
-        raise OSError(errno.EINVAL, 'Invalid argument', None)
-
-    return file_obj
+    try:
+      return FakeFileOpen(self.filesystem).Call(file_des, mode=mode)
+    except IOError as e:
+      raise OSError(e)
 
   def open(self, file_path, flags, mode=None):
     """Returns the file descriptor for a FakeFile.
@@ -1173,7 +1197,7 @@ class FakeOsModule(object):
       OSError: bad file descriptor.
       TypeError: if file descriptor is not an integer.
     """
-    fh = self.fdopen(file_des)
+    fh = self.filesystem.GetOpenFile(file_des)
     fh.close()
 
   def read(self, file_des, num_bytes):
@@ -1190,7 +1214,7 @@ class FakeOsModule(object):
       OSError: bad file descriptor.
       TypeError: if file descriptor is not an integer.
     """
-    fh = self.fdopen(file_des)
+    fh = self.filesystem.GetOpenFile(file_des)
     return fh.read(num_bytes)
 
   def write(self, file_des, contents):
@@ -1207,7 +1231,7 @@ class FakeOsModule(object):
       OSError: bad file descriptor.
       TypeError: if file descriptor is not an integer.
     """
-    fh = self.fdopen(file_des)
+    fh = self.filesystem.GetOpenFile(file_des)
     fh.write(contents)
     fh.flush()
     return len(contents)
@@ -1225,7 +1249,7 @@ class FakeOsModule(object):
       OSError: if the filesystem object doesn't exist.
     """
     # stat should return the tuple representing return value of os.stat
-    stats = self.fdopen(file_des).GetObject()
+    stats = self.filesystem.GetOpenFile(file_des).GetObject()
     st_obj = os.stat_result((stats.st_mode, stats.st_ino, stats.st_dev,
                              stats.st_nlink, stats.st_uid, stats.st_gid,
                              stats.st_size, stats.st_atime,
@@ -1247,8 +1271,8 @@ class FakeOsModule(object):
     """
     try:
       directory = self.filesystem.GetObject(target_directory)
-    except IOError, (errno_code, msg):
-      raise OSError(errno_code, msg, target_directory)
+    except IOError as e:
+      raise OSError(e.errno, e.strerror, target_directory)
     if not directory.st_mode & stat.S_IFDIR:
       raise OSError(errno.ENOTDIR,
                     'Fake os module: not a directory',
@@ -1297,7 +1321,9 @@ class FakeOsModule(object):
     return self.filesystem.cwd
 
   def getcwdu(self):
-    """Return current working directory."""
+    """Return current working directory. Deprecated in Python 3."""
+    if sys.version_info >= (3, 0):
+      raise AttributeError('no attribute getcwdu')
     return unicode(self.filesystem.cwd)
 
   def listdir(self, target_directory):
@@ -1361,7 +1387,7 @@ class FakeOsModule(object):
     top = self.path.normpath(top)
     try:
       top_contents = self._ClassifyDirectoryContents(top)
-    except OSError, e:
+    except OSError as e:
       top_contents = None
       if onerror is not None:
         onerror(e)
@@ -1422,7 +1448,7 @@ class FakeOsModule(object):
                                stats.st_size, stats.st_atime,
                                stats.st_mtime, stats.st_ctime))
       return st_obj
-    except IOError, io_error:
+    except IOError as io_error:
       raise OSError(io_error.errno, io_error.strerror, entry_path)
 
   def lstat(self, entry_path):
@@ -1445,7 +1471,7 @@ class FakeOsModule(object):
                                stats.st_size, stats.st_atime,
                                stats.st_mtime, stats.st_ctime))
       return st_obj
-    except IOError, io_error:
+    except IOError as io_error:
       raise OSError(io_error.errno, io_error.strerror, entry_path)
 
   def remove(self, path):
@@ -1454,7 +1480,7 @@ class FakeOsModule(object):
       raise OSError(errno.EISDIR, "Is a directory: '%s'" % path)
     try:
       self.filesystem.RemoveObject(path)
-    except IOError, e:
+    except IOError as e:
       raise OSError(e.errno, e.strerror, e.filename)
 
   # As per the documentation unlink = remove.
@@ -1524,7 +1550,7 @@ class FakeOsModule(object):
                       target_directory)
       try:
         self.filesystem.RemoveObject(target_directory)
-      except IOError, e:
+      except IOError as e:
         raise OSError(e.errno, e.strerror, e.filename)
 
   def removedirs(self, target_directory):
@@ -1553,7 +1579,7 @@ class FakeOsModule(object):
       dir_name: (str) Name of directory to create.  Relative paths are assumed
         to be relative to '/'.
       mode: (int) Mode to create directory with.  This argument defaults to
-        0777.  The umask is applied to this mode.
+        0o777.  The umask is applied to this mode.
 
     Raises:
       OSError: if the directory name is invalid or parent directory is read only
@@ -1587,7 +1613,7 @@ class FakeOsModule(object):
     Args:
       dir_name: (str) Name of directory to create.
       mode: (int) Mode to create directory (and any necessary parent
-        directories) with. This argument defaults to 0777.  The umask is
+        directories) with. This argument defaults to 0o777.  The umask is
         applied to this mode.
 
     Raises:
@@ -1623,7 +1649,7 @@ class FakeOsModule(object):
     """
     try:
       st = self.stat(path)
-    except OSError, os_error:
+    except OSError as os_error:
       if os_error.errno == errno.ENOENT:
         return False
       raise
@@ -1638,7 +1664,7 @@ class FakeOsModule(object):
     """
     try:
       file_object = self.filesystem.GetObject(path)
-    except IOError, io_error:
+    except IOError as io_error:
       if io_error.errno == errno.ENOENT:
         raise OSError(errno.ENOENT,
                       'No such file or directory in fake filesystem',
@@ -1663,7 +1689,7 @@ class FakeOsModule(object):
     """
     try:
       file_object = self.filesystem.GetObject(path)
-    except IOError, io_error:
+    except IOError as io_error:
       if io_error.errno == errno.ENOENT:
         raise OSError(errno.ENOENT,
                       'No such file or directory in fake filesystem',
@@ -1692,7 +1718,7 @@ class FakeOsModule(object):
     """
     try:
       file_object = self.filesystem.GetObject(path)
-    except IOError, io_error:
+    except IOError as io_error:
       if io_error.errno == errno.ENOENT:
         raise OSError(errno.ENOENT,
                       'No such file or directory in fake filesystem',
@@ -1712,7 +1738,7 @@ class FakeOsModule(object):
     Args:
       filename: (str) Name of the file to create
       mode: (int) permissions to use and type of file to be created.
-        Default permissions are 0666.  Only the stat.S_IFREG file type
+        Default permissions are 0o666.  Only the stat.S_IFREG file type
         is supported by the fake implementation.  The umask is applied
         to this mode.
       device: not supported in fake implementation
@@ -1786,14 +1812,35 @@ class FakeFileOpen(object):
     self.filesystem = filesystem
     self._delete_on_close = delete_on_close
 
-  def __call__(self, file_path, flags='r', bufsize=-1):
+  def __call__(self, *args, **kwargs):
+    """Redirects calls to file() or open() to appropriate method."""
+    if sys.version_info < (3, 0):
+      return self._call_ver2(*args, **kwargs)
+    else:
+      return self.Call(*args, **kwargs)
+
+  def _call_ver2(self, file_path, mode='r', buffering=-1, flags=None):
+    """Limits args of open() or file() for Python 2.x versions."""
+    # Backwards compatibility, mode arg used to be named flags
+    mode = flags or mode
+    return self.Call(file_path, mode, buffering)
+
+  def Call(self, file_, mode='r', buffering=-1, encoding=None,
+           errors=None, newline=None, closefd=True, opener=None):
     """Returns a StringIO object with the contents of the target file object.
 
     Args:
-      file_path:  path to target file
-      flags: additional file flags. All r/w/a r+/w+/a+ flags are supported.
-        't', 'b', and 'U' are ignored, e.g., 'wb' is treated as 'w'.
-      bufsize: ignored. (Used for signature compliance with __builtin__.open)
+      file_: path to target file or a file descriptor
+      mode: additional file modes. All r/w/a r+/w+/a+ modes are supported.
+        't', and 'U' are ignored, e.g., 'wU' is treated as 'w'. 'b' sets
+        binary mode, no end of line translations in StringIO.
+      buffering: ignored. (Used for signature compliance with __builtin__.open)
+      encoding: ignored, strings have no encoding
+      errors: ignored, this relates to encoding
+      newline: controls universal newlines, passed to StringIO object
+      closefd: if a file descriptor rather than file name is passed, and set
+        to false, then the file descriptor is kept open when file is closed
+      opener: not supported
 
     Returns:
       a StringIO object containing the contents of the target file
@@ -1802,40 +1849,48 @@ class FakeFileOpen(object):
       IOError: if the target object is a directory, the path is invalid or
         permission is denied.
     """
-    orig_flags = flags  # Save original flags for error messages.
-    # Normalize flags. Ignore 't', 'b', and 'U'.
-    flags = flags.replace('t', '').replace('b', '')
-    flags = flags.replace('rU', 'r').replace('U', 'r')
+    orig_modes = mode  # Save original mdoes for error messages.
+    # Binary mode for non 3.x or set by mode
+    binary = sys.version_info < (3, 0) or 'b' in mode
+    # Normalize modes. Ignore 't' and 'U'.
+    mode = mode.replace('t', '').replace('b', '')
+    mode = mode.replace('rU', 'r').replace('U', 'r')
 
-    if flags not in _OPEN_MODE_MAP:
-      raise IOError('Invalid mode: %r' % orig_flags)
+    if mode not in _OPEN_MODE_MAP:
+      raise IOError('Invalid mode: %r' % orig_modes)
 
-    must_exist, need_read, need_write, truncate, append = _OPEN_MODE_MAP[flags]
+    must_exist, need_read, need_write, truncate, append = _OPEN_MODE_MAP[mode]
 
-    real_path = self.filesystem.ResolvePath(file_path)
-    if self.filesystem.Exists(file_path):
-      file_object = self.filesystem.GetObjectFromNormalizedPath(real_path)
+    file_object = None
+    filedes = None
+    # opening a file descriptor
+    if isinstance(file_, int):
+      filedes = file_
+      file_object = self.filesystem.GetOpenFile(filedes).GetObject()
+      file_path = file_object.name
+    else:
+      file_path = file_
+      real_path = self.filesystem.ResolvePath(file_path)
+      if self.filesystem.Exists(file_path):
+        file_object = self.filesystem.GetObjectFromNormalizedPath(real_path)
+      closefd = True
+
+    if file_object:
       if ((need_read and not file_object.st_mode & PERM_READ) or
           (need_write and not file_object.st_mode & PERM_WRITE)):
-        raise IOError(errno.EACCES,
-                      'Permission denied',
-                      file_path)
+        raise IOError(errno.EACCES, 'Permission denied', file_path)
       if need_write:
         file_object.st_ctime = int(time.time())
         if truncate:
           file_object.SetContents('')
     else:
       if must_exist:
-        raise IOError(errno.ENOENT,
-                      'No such file or directory',
-                      file_path)
+        raise IOError(errno.ENOENT, 'No such file or directory', file_path)
       file_object = self.filesystem.CreateFile(
           real_path, create_missing_dirs=False, apply_umask=True)
 
     if file_object.st_mode & stat.S_IFDIR:
-      raise IOError(errno.EISDIR,
-                    'Fake file object: is a directory',
-                    file_path)
+      raise IOError(errno.EISDIR, 'Fake file object: is a directory', file_path)
 
     class FakeFileWrapper(object):
       """Wrapper for a StringIO object for use by a FakeFile object.
@@ -1845,14 +1900,17 @@ class FakeFileOpen(object):
       """
 
       def __init__(self, file_object, update=False, read=False, append=False,
-                   delete_on_close=False, filesystem=None):
+                   delete_on_close=False, filesystem=None, newline=None,
+                   binary=True, closefd=True):
         self._file_object = file_object
         self._append = append
         self._read = read
         self._update = update
+        self._closefd = closefd
+        newline_arg = {} if binary else {'newline': newline}
         if file_object.contents:
           if update:
-            self._io = cStringIO.StringIO()
+            self._io = io.StringIO(**newline_arg)
             self._io.write(file_object.contents)
             if not append:
               self._io.seek(0)
@@ -1863,9 +1921,9 @@ class FakeFileOpen(object):
               else:
                 self._read_seek = self._io.tell()
           else:
-            self._io = cStringIO.StringIO(file_object.contents)
+            self._io = io.StringIO(file_object.contents, **newline_arg)
         else:
-          self._io = cStringIO.StringIO()
+          self._io = io.StringIO(**newline_arg)
           self._read_whence = 0
           self._read_seek = 0
         if delete_on_close:
@@ -1880,8 +1938,7 @@ class FakeFileOpen(object):
         """To support usage of this fake file with the 'with' statement."""
         return self
 
-      # pylint: disable-msg=W0622
-      def __exit__(self, type, value, traceback):
+      def __exit__(self, type, value, traceback):  # pylint: disable-msg=W0622
         """To support usage of this fake file with the 'with' statement."""
         self.close()
 
@@ -1897,7 +1954,8 @@ class FakeFileOpen(object):
         """File close."""
         if self._update:
           self._file_object.SetContents(self._io.getvalue())
-        self._filesystem.CloseOpenFile(self)
+        if self._closefd:
+          self._filesystem.CloseOpenFile(self)
         if self._delete_on_close:
           self._filesystem.RemoveObject(self.name)
 
@@ -1963,8 +2021,8 @@ class FakeFileOpen(object):
           while we're in append mode goes through this.
 
           Args:
-            args: pass through args
-            kwargs: pass through kwargs
+            *args: pass through args
+            **kwargs: pass through kwargs
           Returns:
             Wrapped StringIO object method
           """
@@ -1995,8 +2053,8 @@ class FakeFileOpen(object):
           the read pointer as well.
 
           Args:
-            args: pass through args
-            kwargs: pass through kwargs
+            *args: pass through args
+            **kwargs: pass through kwargs
           Returns:
             Wrapped StringIO object method
           """
@@ -2035,8 +2093,14 @@ class FakeFileOpen(object):
                                read=need_read,
                                append=append,
                                delete_on_close=self._delete_on_close,
-                               filesystem=self.filesystem)
-    fakefile.filedes = self.filesystem.AddOpenFile(fakefile)
+                               filesystem=self.filesystem,
+                               newline=newline,
+                               binary=binary,
+                               closefd=closefd)
+    if filedes is not None:
+      fakefile.filedes = filedes
+    else:
+      fakefile.filedes = self.filesystem.AddOpenFile(fakefile)
     return fakefile
 
 
